#include <iostream>
#include <cstdio>
#include <cstring>
#include <thread>
#include <mutex>
#include <condition_variable>
#include <atomic>

using namespace std;

#define BUFFER_DIM 1024
#define BUFFER_SIZE 64

struct BUFFER
{
 unsigned char buffer[BUFFER_DIM];
 int n;
};

class thread_safe_buffer
{
 private:

 BUFFER ring_buffer[BUFFER_SIZE];
 int read_index;
 int write_index;
 atomic<int> N_block;
 mutex producer, consumer;
 condition_variable_any not_full, not_empty;

 public:

 // costruttore
 thread_safe_buffer(void)
 {
  read_index = 0;
  write_index = 0;
  N_block = 0;
 }
 
 // acquisizione contatore numero di blocchi dati
 int getNblock(void)
 {
  return N_block;
 }

 // inserimento di un blocco dati nel buffer circolare
 // (con attesa in caso di buffer pieno)
 void put(unsigned char data_block[], int data_block_dim)
 {
  int dim;
  
  producer.lock();
  // eventuale attesa della condizione di buffer non pieno
  if (N_block >= BUFFER_SIZE)
    not_full.wait(producer);
  // limitazione della dimensione del blocco dati
  if (data_block_dim < BUFFER_DIM)
    dim = data_block_dim;
  else
      dim = BUFFER_DIM;
  // copia del blocco dati nel buffer circolare
  memcpy(ring_buffer[write_index].buffer, data_block, dim); // put
  ring_buffer[write_index].n = dim;
  // aggiornamento indice di scrittura e contatore numero di blocchi
  write_index = (write_index + 1) % BUFFER_SIZE;
  N_block++;
  // notifica condizione di buffer non vuoto
  not_empty.notify_one();
  producer.unlock();
}

 // estrazione di un blocco dati dal buffer circolare
 // (con attesa in caso di buffer vuoto)
 // restituisce la dimensione del blocco dati
 int get(unsigned char *data_block)
 {
  int dim;
  
  consumer.lock();
  // eventuale attesa della condizione di buffer non vuoto
  if (N_block == 0)
    not_empty.wait(consumer);
  dim = ring_buffer[read_index].n;
  // copia del blocco dati dal buffer circolare
  memcpy( data_block, ring_buffer[read_index].buffer, dim); // get
  // aggiornamento indice di lettura e contatore numero di blocchi
  read_index = (read_index + 1) % BUFFER_SIZE;
  N_block--;
  // notifica condizione di buffer non pieno
  not_full.notify_one();
  consumer.unlock();
  return dim;
 }
};


thread_safe_buffer data_block_buffer;
bool eof = false;

// lettura dei dati da file
void leggi(FILE* file)
{
 unsigned char buffer[BUFFER_DIM];

 while (!feof(file))
      {
       int n = fread(buffer, 1, BUFFER_DIM, file);
       if (n > 0)
         data_block_buffer.put(buffer, n);
       }
 eof = true;
}

// scrittura dei dati su file
void scrivi(FILE* file)
{
 unsigned char buffer[BUFFER_DIM];

 while (!eof || data_block_buffer.getNblock()>0)
      {
       int n = data_block_buffer.get(buffer);
       fwrite(buffer, 1, n, file);
      }
}




int main(int argc, char*argv[])
{
 FILE *origine, *destinazione;

 // verifica degli argomenti della riga di comando
 if (argc != 3)
   {
    cout << "Uso: " << argv[0] << " file-origine file-destinazione" << endl;
    return -1;
   }
 // apertura in lettura del file di origine
 origine = fopen(argv[1], "rb");
 if (origine == NULL)
   {
    cout << "Errore apertura file" << argv[1] << endl;
    return -1;
   }
 destinazione = fopen(argv[2], "wb");
 // apertura in scrittura del file di destinazione
 if (destinazione == NULL)
   {
    cout << "Errore apertura file" << argv[2] << endl;
    fclose(origine);
    return -1;
   }
 // esecuzione dei thread di lettura e scrittura
 thread lettura(leggi, origine);
 thread scrittura(scrivi, destinazione);
 // attesa della conclusione dei thread e chiusura dei file
 lettura.join();
 scrittura.join();
 fclose(origine);
 fclose(destinazione);
 return 0;
}
